#!/usr/bin/env python
"""This module defines an ASE interface to XTB 6.2.0 through I/O
because XTB 6.3.3 seems to have some issues. """

import re
import os
import numpy as np

import ase
import ase.io.turbomole as io
from ase.calculators.calculator import FileIOCalculator, Parameters, ReadError
from ase.constraints import FixAtoms
from ase.units import Bohr, Hartree
from subprocess import check_output


def get_xtb_version(command):
    mess = "Output from version check:\n"
    for line in check_output(command, shell=True).decode("ascii").split("\n"):
        if "Version" in line or "version" in line:
            check = True
        else:
            mess += line + "\n"
    if not check:
        raise RuntimeError("Cannot recognize XTB version.\n {0}".format(mess))
    return line


class XTB(FileIOCalculator):

    implemented_properties = ["energy", "forces"]  # TO ADD MORE PROPERTIES IF NEEDED
    discard_results_on_any_change = True

    def __init__(
        self,
        restart=None,
        ignore_bad_restart_file=False,  # UPDATE TO ignore_bad_restarted_file=FileIOCalculator._deprecated SOON
        label="xtbcalc/",
        atoms=None,
        command="/home/afabrizi/MyQuantumChemSoftwares/xtb/bin/xtb --gfn 0 --grad"
        + " xtb.tmol > xtb.log",
        verb=0,
    ):
        xtbpath = os.environ.get("XTBPATH")
        if xtbpath is None:
            mess = "The XTBPATH variable is not defined."
            raise ValueError(mess)
        self.label = label
        self.verb = verb  # VERBOSITY FLAG
        FileIOCalculator.__init__(
            self,
            restart=restart,
            ignore_bad_restart_file=ignore_bad_restart_file,
            label=label,
            atoms=atoms,
            command=command,
        )
        self.clean()
        get_xtb_version(
            "/home/afabrizi/MyQuantumChemSoftwares/xtb/bin/xtb --version"
        )  # SIMPLE CHECK, CAN BE CUSTOMIZED TO GRAB/PRINT VERSION NUMBER

    def write_charge_mult(self):
        charge = int(self.atoms.get_initial_charges().sum().round())
        uhf = int(self.atoms.get_initial_magnetic_moments().sum().round())
        with open(self.label + ".CHRG", "w") as f1:
            f1.write(str(charge))
        with open(self.label + ".UHF", "w") as f2:
            f2.write(str(uhf))

    def write_input(self, atoms, properties=None, system_changes=None):
        if not (np.all(atoms.pbc) or not np.any(atoms.pbc)):
            raise RuntimeError("PBC must be all true or all false")
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        filename = self.label + "xtb.tmol"
        coord = atoms.get_positions()
        symbols = atoms.get_chemical_symbols()
        self.write_charge_mult()
        with open(filename, "w") as fd:
            fix_indices = set()
            if self.atoms.constraints:
                for constr in atoms.constraints:
                    if isinstance(constr, FixAtoms):
                        fix_indices.update(constr.get_indices())

            fix_str = []
            for i in range(len(atoms)):
                if i in fix_indices:
                    fix_str.append("f")
                else:
                    fix_str.append("")
            if np.all(atoms.pbc):
                fd.write("$periodic 3\n")
                fd.write("$cell angs\n")
                fd.write(
                    " {0} {1} {2}  ".format(
                        atoms.get_cell_lengths_and_angles()[0],
                        atoms.get_cell_lengths_and_angles()[1],
                        atoms.get_cell_lengths_and_angles()[2],
                    )
                )
                fd.write(
                    " {0} {1} {2}\n".format(
                        atoms.get_cell_lengths_and_angles()[3],
                        atoms.get_cell_lengths_and_angles()[4],
                        atoms.get_cell_lengths_and_angles()[5],
                    )
                )
            fd.write("$coord angs\n")
            for (x, y, z), s, fix in zip(coord, symbols, fix_str):
                fd.write(
                    "%20.14f  %20.14f  %20.14f      %2s  %2s \n"
                    % (x, y, z, s.lower(), fix)
                )
            fd.write("$end\n")

    def read_results(self, verb=None):
        if verb is None:
            verb = self.verb
        self.read_energy(self.directory, self.label, verb)
        self.read_forces(self.atoms, self.directory, self.label, verb)
        #self.clean()
        # TO ADD MORE PROPERTIES IF NEEDED

    def read_forces(self, atoms, directory, label, verb=0):
        gradient_file = label + "gradient"
        natoms = len(atoms)
        if not os.path.exists(gradient_file):
            raise ReadError("The gradient file does not exist.")
        if os.path.isfile(gradient_file):
            with open(gradient_file, "r") as fd:
                lines = fd.readlines()
        lines = lines[-1 - natoms : -1]
        assert len(lines) == natoms
        self.results["forces"] = np.zeros((natoms, 3), float)
        for i in range(natoms):
            line = [s for s in lines[i].strip().split(" ") if len(s) > 0]
            f = -np.array([float(x.replace("D", "E")) for x in line[0:3]])
            self.results["forces"][i, :] = f * (Hartree / Bohr)
        os.remove(gradient_file)
        if verb > 1:
            print("Forces of the last atom:\n {0}".format(f * (Hartree / Bohr)))

    def read_energy(self, directory, label, verb=0):
        energy_file = label + "energy"
        if not os.path.exists(energy_file):
            raise ReadError("The energy file does not exist.")
        if os.path.isfile(energy_file):
            with open(energy_file, "r") as fd:
                lines = fd.readlines()
        for i in range(len(lines)):
            if lines[i].startswith("$end"):
                escf = (
                    float(lines[i - 1].strip().split(" ")[3].replace("D", "E"))
                    * Hartree
                )
                self.results["energy"] = escf
                os.remove(energy_file)
                break
        if verb > 0:
            print("Energy:\n {0}".format(escf))

    def finished_successfully(self):
        finished = False
        mess = ""
        for line in open(self.label + "xtb.log", "r"):
            if line.rfind("finished run") > -1:
                finished = True
            if line.contains("error"):
                mess += line
        if verb > 0:
            if finished:
                print("Normal termination.")
            else:
                print("Error termination.")
        return finished, mess

    def clean(self):
        files_to_clean = [
            self.label + "xtb.log",
            self.label + "energy",
            self.label + "gradient",
            self.label + "xtbrestart",
            self.label + "charges",
        ]  # POSSIBLE MORE STUFF TO DELETE BETWEEN EXECUTIONS
        for f in files_to_clean:
            try:
                os.remove(f)
            except OSError:
                pass


if __name__ == "__main__":
    from ase.build import molecule

    atoms = molecule("O2")
    atoms.set_calculator(XTB())
    print("Potential energy %5.5f eV" % atoms.get_potential_energy())
    print(
        "Forces per atom:\n {0} \n {1}".format(
            atoms.get_forces()[0][:], atoms.get_forces()[1][:]
        )
    )


# Hope it works!
